#include "RandomUpdateCullingManager.h"
#include <time.h>
#include "CullingLogManager.h"


namespace GtpVisibility {

const static int R_CANDIDATES = 1;
//-----------------------------------------------------------------------
RandomUpdateCullingManager::RandomUpdateCullingManager()
{
	SetAssumedVisibility(0);
	// initialise random generator in case we use assumed visibility
	srand(time(NULL));

	mRandomCandidates = R_CANDIDATES;
}
//-----------------------------------------------------------------------
RandomUpdateCullingManager::RandomUpdateCullingManager(
			const unsigned int assumedVisibility)
{
	SetAssumedVisibility(assumedVisibility);
	// initialise random generator in case we use assumed visibility
	srand(time(NULL));

	mRandomCandidates = R_CANDIDATES;
}
//-----------------------------------------------------------------------
void RandomUpdateCullingManager::RenderScene()
{
	CullingLogManager::GetSingleton()->LogMessage("ruc");

	QueryQueue queryQueue;
	unsigned int visiblePixels = 0;
	
	/////////////
	//-- PART 1: process finished occlusion queries

	// update the fully visible classifications
	mHierarchyInterface->DetermineFullVisibility(mHierarchyInterface->GetHierarchyRoot());

	//mHierarchyInterface->TraverseNode2(mHierarchyInterface->GetHierarchyRoot());
	
#if 1
	while (!mHierarchyInterface->GetQueue()->empty() || !queryQueue.empty())
	{
		bool resultAvailable = false;

		//-- only wait for result if there are no nodes to process
		while (!queryQueue.empty() && 
				queryQueue.front().second->GetQueryResult(visiblePixels, 
				mHierarchyInterface->GetQueue()->empty()))
		{
        	HierarchyNode *node = queryQueue.front().first;
			
			queryQueue.pop();
			
			if (visiblePixels > mVisibilityThreshold)
			{
				// in case geometry is in interior node: 
				// ensure that we only traverse once
				if (!mHierarchyInterface->IsNodeVisible(node) &&
					!mHierarchyInterface->IsNodeFullyVisible(node))
				{
					mHierarchyInterface->TraverseNode(node);
				}

				mHierarchyInterface->PullUpVisibility(node);
			}
			else
			{	
				mHierarchyInterface->SetNodeVisible(node, false);

				++ mNumQueryCulledNodes;
				
				if (mVisualizeCulledNodes)
				{
					mHierarchyInterface->VisualizeCulledNode(node, QUERY_CULLED);
				}
			}
		}
		
		//-- PART 2: hierarchical traversal
		if (!mHierarchyInterface->GetQueue()->empty())
		{
			HierarchyNode *node = mHierarchyInterface->GetQueue()->top();
			mHierarchyInterface->GetQueue()->pop();
				
			bool intersects = false;

			if (!mHierarchyInterface->CheckFrustumVisible(node, intersects))
			{
				++ mNumFrustumCulledNodes;

				if (mVisualizeCulledNodes)
				{
					mHierarchyInterface->VisualizeCulledNode(node, FRUSTUM_CULLED);
				}
			}
			//-- if node intersects near plane, skip query because wrong results possible
			else if (intersects)
			{
				SkipQuery(node);
			}
			else
			{
				// identify previously visible nodes
				const bool wasVisible = mHierarchyInterface->IsNodeVisible(node) && 
					(mHierarchyInterface->LastVisited(node) == mHierarchyInterface->GetFrameId() - 1);
				
				// if we assume node to be visible in this frame => skip query 
				const bool skipQuery = wasVisible && (mAssumedVisibility > 0) && 
					DecideVisible(node) && mHierarchyInterface->HasGeometry(node);

				if (skipQuery)
				{
					SkipQuery(node);
					continue;
				}
								
				if (mHierarchyInterface->IsNodeFullyVisible(node))
				{
					// node fully visible => test only random leaves
					for (int i = 0; i < mRandomCandidates; ++ i)
					{
						HierarchyNode *randomLeaf = mHierarchyInterface->GetRandomLeaf(node);
						
						if (!randomLeaf)
							continue;

						mHierarchyInterface->SetNodeVisible(randomLeaf, false);

						// update node's visited flag
						mHierarchyInterface->SetLastVisited(randomLeaf, 
															mHierarchyInterface->GetFrameId());

						queryQueue.push(QueryPair(randomLeaf, 
							mHierarchyInterface->IssueNodeOcclusionQuery(node)));
					}

					mHierarchyInterface->TraverseNode2(node);

					continue;
				}

                // identify nodes that we cannot skip queries for
				// geometry not only in leaves => test for renderable geometry
				const bool issueQuery = !wasVisible || mHierarchyInterface->HasGeometry(node);
							
				// reset node's visibility classification
				// set visibe if geometry in node => we only traverse the node once
				mHierarchyInterface->SetNodeVisible(node, wasVisible && issueQuery);

				// update node's visited flag
				mHierarchyInterface->SetLastVisited(node, mHierarchyInterface->GetFrameId());
				

				// skip testing previously visible nodes without geometry
				if (issueQuery)
				{
					++ mNumQueriesIssued;
					
					queryQueue.push(QueryPair(node, mHierarchyInterface->
						IssueNodeOcclusionQuery(node, wasVisible)));
				}
				
				// always traverse a node if it was visible
				if (wasVisible)
				{
					mHierarchyInterface->TraverseNode(node);
				}
			}
		}
	}
#endif
	//while (!mHierarchyInterface->GetQueue()->empty())
	//	mHierarchyInterface->GetQueue()->pop();
	
}
//-----------------------------------------------------------------------
void RandomUpdateCullingManager::SetAssumedVisibility(const unsigned int assumedVisibility)
{
	mAssumedVisibility = assumedVisibility;
	
	mThreshold = 0;

	if (mAssumedVisibility > 0)
	{
		mThreshold = RAND_MAX - RAND_MAX / mAssumedVisibility;
		if (mAssumedVisibility > 100) // fix visibility
			mThreshold = RAND_MAX;
	}

}
//-----------------------------------------------------------------------
inline bool RandomUpdateCullingManager::DecideVisible(HierarchyNode *node) const
{
	return rand() < mThreshold;
}
//-----------------------------------------------------------------------
inline void RandomUpdateCullingManager::SkipQuery(HierarchyNode *node) const
{
	// -- set node to be visible in this frame, then traverse it
	mHierarchyInterface->SetLastVisited(node, mHierarchyInterface->GetFrameId());
	
	mHierarchyInterface->PullUpVisibility(node);			
	mHierarchyInterface->TraverseNode(node);
}

} // namespace GtpVisibility
